@@ -464,7 +464,7 @@ cloneInitiate (CompDisplay     *d,
 
 	CLONE_SCREEN (s);
 
-	if (cs->grab || otherScreenGrabExist (s, "clone", 0))
+	if (cs->grab || otherScreenGrabExist (s, "clone", NULL))
 	    return FALSE;
 
 	if (!cs->grabIndex)
@@ -864,8 +864,9 @@ cloneGetObjectOptions (CompPlugin *plugin,
 	(GetPluginObjectOptionsProc) cloneGetDisplayOptions
     };
 
+    *count = 0;
     RETURN_DISPATCH (object, dispTab, ARRAY_SIZE (dispTab),
-		     (void *) (*count = 0), (plugin, object, count));
+		     (void *) count, (plugin, object, count));
 }
 
 static CompBool
