@@ -179,50 +179,6 @@ closeWin (CompDisplay     *d,
     return TRUE;
 }
 
-static Bool
-mainMenu (CompDisplay     *d,
-	  CompAction      *action,
-	  CompActionState state,
-	  CompOption      *option,
-	  int		  nOption)
-{
-    CompScreen   *s;
-    Window       xid;
-    unsigned int time;
-
-    xid  = getIntOptionNamed (option, nOption, "root", 0);
-    time = getIntOptionNamed (option, nOption, "time", CurrentTime);
-
-    s = findScreenAtDisplay (d, xid);
-    if (s && !s->maxGrab)
-	toolkitAction (s, s->display->toolkitActionMainMenuAtom, time, s->root,
-		       0, 0, 0);
-
-    return TRUE;
-}
-
-static Bool
-runDialog (CompDisplay     *d,
-	   CompAction      *action,
-	   CompActionState state,
-	   CompOption      *option,
-	   int		   nOption)
-{
-    CompScreen   *s;
-    Window       xid;
-    unsigned int time;
-
-    xid  = getIntOptionNamed (option, nOption, "root", 0);
-    time = getIntOptionNamed (option, nOption, "time", CurrentTime);
-
-    s = findScreenAtDisplay (d, xid);
-    if (s && !s->maxGrab)
-	toolkitAction (s, s->display->toolkitActionRunDialogAtom, time, s->root,
-		       0, 0, 0);
-
-    return TRUE;
-}
-
 static Bool
 unmaximize (CompDisplay     *d,
 	    CompAction      *action,
@@ -399,100 +355,6 @@ lowerInitiate (CompDisplay     *d,
     return TRUE;
 }
 
-static Bool
-runCommandDispatch (CompDisplay     *d,
-		    CompAction      *action,
-		    CompActionState state,
-		    CompOption      *option,
-		    int		    nOption)
-{
-    CompScreen *s;
-    Window     xid;
-
-    xid = getIntOptionNamed (option, nOption, "root", 0);
-
-    s = findScreenAtDisplay (d, xid);
-    if (s)
-    {
-	int index = -1;
-	int i = COMP_DISPLAY_OPTION_RUN_COMMAND0_KEY;
-
-	while (i <= COMP_DISPLAY_OPTION_RUN_COMMAND11_KEY)
-	{
-	    if (action == &d->opt[i].value.action)
-	    {
-		index = i - COMP_DISPLAY_OPTION_RUN_COMMAND0_KEY +
-		    COMP_DISPLAY_OPTION_COMMAND0;
-		break;
-	    }
-
-	    i++;
-	}
-
-	if (index > 0)
-	    runCommand (s, d->opt[index].value.s);
-    }
-
-    return TRUE;
-}
-
-static Bool
-runCommandScreenshot (CompDisplay     *d,
-		      CompAction      *action,
-		      CompActionState state,
-		      CompOption      *option,
-		      int	      nOption)
-{
-    CompScreen *s;
-    Window     xid;
-
-    xid = getIntOptionNamed (option, nOption, "root", 0);
-
-    s = findScreenAtDisplay (d, xid);
-    if (s)
-	runCommand (s, d->opt[COMP_DISPLAY_OPTION_SCREENSHOT].value.s);
-
-    return TRUE;
-}
-
-static Bool
-runCommandWindowScreenshot (CompDisplay     *d,
-			    CompAction      *action,
-			    CompActionState state,
-			    CompOption      *option,
-			    int	            nOption)
-{
-    CompScreen *s;
-    Window     xid;
-
-    xid = getIntOptionNamed (option, nOption, "root", 0);
-
-    s = findScreenAtDisplay (d, xid);
-    if (s)
-	runCommand (s, d->opt[COMP_DISPLAY_OPTION_WINDOW_SCREENSHOT].value.s);
-
-    return TRUE;
-}
-
-static Bool
-runCommandTerminal (CompDisplay     *d,
-		    CompAction      *action,
-		    CompActionState state,
-		    CompOption      *option,
-		    int	            nOption)
-{
-    CompScreen *s;
-    Window     xid;
-
-    xid = getIntOptionNamed (option, nOption, "root", 0);
-
-    s = findScreenAtDisplay (d, xid);
-    if (s)
-	runCommand (s, d->opt[COMP_DISPLAY_OPTION_TERMINAL].value.s);
-
-    return TRUE;
-}
-
 static Bool
 windowMenu (CompDisplay     *d,
 	    CompAction      *action,
@@ -621,32 +483,6 @@ const CompMetadataOptionInfo coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM] = {
     { "autoraise_delay", "int", 0, 0, 0 },
     { "close_window_key", "key", 0, closeWin, 0 },
     { "close_window_button", "button", 0, closeWin, 0 },
-    { "main_menu_key", "key", 0, mainMenu, 0 },
-    { "run_key", "key", 0, runDialog, 0 },
-    { "command0", "string", 0, 0, 0 },
-    { "command1", "string", 0, 0, 0 },
-    { "command2", "string", 0, 0, 0 },
-    { "command3", "string", 0, 0, 0 },
-    { "command4", "string", 0, 0, 0 },
-    { "command5", "string", 0, 0, 0 },
-    { "command6", "string", 0, 0, 0 },
-    { "command7", "string", 0, 0, 0 },
-    { "command8", "string", 0, 0, 0 },
-    { "command9", "string", 0, 0, 0 },
-    { "command10", "string", 0, 0, 0 },
-    { "command11", "string", 0, 0, 0 },
-    { "run_command0_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command1_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command2_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command3_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command4_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command5_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command6_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command7_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command8_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command9_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command10_key", "key", 0, runCommandDispatch, 0 },
-    { "run_command11_key", "key", 0, runCommandDispatch, 0 },
     { "slow_animations_key", "key", 0, toggleSlowAnimations, 0 },
     { "raise_window_key", "key", 0, raiseInitiate, 0 },
     { "raise_window_button", "button", 0, raiseInitiate, 0 },
@@ -658,11 +494,6 @@ const CompMetadataOptionInfo coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM] = {
     { "maximize_window_key", "key", 0, maximize, 0 },
     { "maximize_window_horizontally_key", "key", 0, maximizeHorizontally, 0 },
     { "maximize_window_vertically_key", "key", 0, maximizeVertically, 0 },
-    { "command_screenshot", "string", 0, 0, 0 },
-    { "run_command_screenshot_key", "key", 0, runCommandScreenshot, 0 },
-    { "command_window_screenshot", "string", 0, 0, 0 },
-    { "run_command_window_screenshot_key", "key", 0,
-      runCommandWindowScreenshot, 0 },
     { "window_menu_button", "button", 0, windowMenu, 0 },
     { "window_menu_key", "key", 0, windowMenu, 0 },
     { "show_desktop_key", "key", 0, showDesktop, 0 },
@@ -678,8 +509,6 @@ const CompMetadataOptionInfo coreDisplayOptionInfo[COMP_DISPLAY_OPTION_NUM] = {
     { "hide_skip_taskbar_windows", "bool", 0, 0, 0 },
     { "toggle_window_shaded_key", "key", 0, shade, 0 },
     { "ignore_hints_when_maximized", "bool", 0, 0, 0 },
-    { "command_terminal", "string", 0, 0, 0 },
-    { "run_command_terminal_key", "key", 0, runCommandTerminal, 0 },
     { "ping_delay", "int", "<min>1000</min>", 0, 0 },
     { "edge_delay", "int", "<min>0</min>", 0, 0 }
 };
@@ -844,26 +673,81 @@ setDisplayOption (CompPlugin	  *plugin,
 static void
 updatePlugins (CompDisplay *d)
 {
-    CompOption *o;
-    CompPlugin *p, **pop = 0;
-    int	       nPop, i, j;
+    CompOption      *o;
+    CompPlugin      *p, **pop = 0;
+    int	            nPop, i, j, k;
+    CompOptionValue *pList;
+    int             pListCount = 1;
 
     d->dirtyPluginList = FALSE;
 
     o = &d->opt[COMP_DISPLAY_OPTION_ACTIVE_PLUGINS];
 
-    /* The old plugin list always begins with the core plugin. To make sure
-       we don't unnecessarily unload plugins if the new plugin list does not
-       contain the core plugin, we have to use an offset */
-    if (o->value.list.nValue > 0 && strcmp (o->value.list.value[0].s, "core"))
-	i = 0;
-    else
-	i = 1;
+    /* determine number of plugins, which is core + initial plugins +
+       plugins in option list additional to initial plugins */
+    for (i = 0; i < nInitialPlugins; i++)
+	if (strcmp (initialPlugins[i], "core") != 0)
+	    pListCount++;
+
+    for (i = 0; i < o->value.list.nValue; i++)
+    {
+	if (strcmp (o->value.list.value[i].s, "core") == 0)
+	    continue;
+
+	for (j = 0; j < nInitialPlugins; j++)
+	{
+	    if (strcmp (o->value.list.value[i].s, initialPlugins[j]) == 0)
+		break;
+	}
+
+	/* plugin is not in initial plugin list */
+	if (j == nInitialPlugins)
+	    pListCount++;
+    }
+
+    pList = malloc (sizeof (CompOptionValue) * pListCount);
+    if (!pList)
+    {
+	(*core.setOptionForPlugin) (&d->base, "core", o->name, &d->plugin);
+	return;
+    }
+
+    /* new plugin list needs core as first plugin */
+    pList[0].s = "core";
+
+    /* afterwards, add the initial plugins */
+    for (j = 0, k = 1; j < nInitialPlugins; j++)
+    {
+	/* avoid adding core twice */
+	if (strcmp (initialPlugins[j], "core") == 0)
+	    continue;
+	
+	pList[k++].s = initialPlugins[j];
+    }
+    j = k;
+
+    /* then add the plugins not in the initial plugin list */
+    for (i = 0; i < o->value.list.nValue; i++)
+    {
+	if (strcmp (o->value.list.value[i].s, "core") == 0)
+	    continue;
+
+	for (k = 0; k < nInitialPlugins; k++)
+	{
+	    if (strcmp (o->value.list.value[i].s, initialPlugins[k]) == 0)
+		break;
+	}
+	
+	if (k == nInitialPlugins)
+	    pList[j++].s = o->value.list.value[i].s;
+    }
+
+    assert (j == pListCount);
 
     /* j is initialized to 1 to make sure we never pop the core plugin */
-    for (j = 1; j < d->plugin.list.nValue && i < o->value.list.nValue; i++, j++)
+    for (i = j = 1; j < d->plugin.list.nValue && i < pListCount; i++, j++)
     {
-	if (strcmp (d->plugin.list.value[j].s, o->value.list.value[i].s))
+	if (strcmp (d->plugin.list.value[j].s, pList[i].s))
 	    break;
     }
 
@@ -875,6 +759,7 @@ updatePlugins (CompDisplay *d)
 	if (!pop)
 	{
 	    (*core.setOptionForPlugin) (&d->base, "core", o->name, &d->plugin);
+	    free (pList);
 	    return;
 	}
     }
@@ -886,13 +771,12 @@ updatePlugins (CompDisplay *d)
 	free (d->plugin.list.value[d->plugin.list.nValue].s);
     }
 
-    for (; i < o->value.list.nValue; i++)
+    for (; i < pListCount; i++)
     {
 	p = 0;
 	for (j = 0; j < nPop; j++)
 	{
-	    if (pop[j] && strcmp (pop[j]->vTable->name,
-				  o->value.list.value[i].s) == 0)
+	    if (pop[j] && strcmp (pop[j]->vTable->name, pList[i].s) == 0)
 	    {
 		if (pushPlugin (pop[j]))
 		{
@@ -905,7 +789,7 @@ updatePlugins (CompDisplay *d)
 
 	if (p == 0)
 	{
-	    p = loadPlugin (o->value.list.value[i].s);
+	    p = loadPlugin (pList[i].s);
 	    if (p)
 	    {
 		if (!pushPlugin (p))
@@ -946,6 +830,7 @@ updatePlugins (CompDisplay *d)
     if (nPop)
 	free (pop);
 
+    free (pList);
     (*core.setOptionForPlugin) (&d->base, "core", o->name, &d->plugin);
 }
 
@@ -1914,6 +1799,46 @@ freeDisplay (CompDisplay *d)
     free (d);
 }
 
+static Bool
+aquireSelection (CompDisplay *d,
+		 int         screen,
+		 const char  *name,
+		 Atom        selection,
+		 Window      owner,
+		 Time        timestamp)
+{
+    Display *dpy = d->display;
+    Window  root = XRootWindow (dpy, screen);
+    XEvent  event;
+
+    XSetSelectionOwner (dpy, selection, owner, timestamp);
+
+    if (XGetSelectionOwner (dpy, selection) != owner)
+    {
+	compLogMessage ("core", CompLogLevelError,
+			"Could not acquire %s manager "
+			"selection on screen %d display \"%s\"",
+			name, screen, DisplayString (dpy));
+
+	return FALSE;
+    }
+
+    /* Send client message indicating that we are now the manager */
+    event.xclient.type         = ClientMessage;
+    event.xclient.window       = root;
+    event.xclient.message_type = d->managerAtom;
+    event.xclient.format       = 32;
+    event.xclient.data.l[0]    = timestamp;
+    event.xclient.data.l[1]    = selection;
+    event.xclient.data.l[2]    = 0;
+    event.xclient.data.l[3]    = 0;
+    event.xclient.data.l[4]    = 0;
+
+    XSendEvent (dpy, root, FALSE, StructureNotifyMask, &event);
+
+    return TRUE;
+}
+
 Bool
 addDisplay (const char *name)
 {
@@ -1998,8 +1923,6 @@ addDisplay (const char *name)
 	return FALSE;
     }
 
-    d->connection = XGetXCBConnection (dpy);
-
     if (!compInitDisplayOptionsFromMetadata (d,
 					     &coreMetadata,
 					     coreDisplayOptionInfo,
@@ -2154,11 +2077,14 @@ addDisplay (const char *name)
     d->wmDeleteWindowAtom = XInternAtom (dpy, "WM_DELETE_WINDOW", 0);
     d->wmTakeFocusAtom	  = XInternAtom (dpy, "WM_TAKE_FOCUS", 0);
     d->wmPingAtom	  = XInternAtom (dpy, "_NET_WM_PING", 0);
-    d->wmSyncRequestAtom  = XInternAtom (dpy, "_NET_WM_SYNC_REQUEST", 0);
 
+    d->wmSyncRequestAtom  = XInternAtom (dpy, "_NET_WM_SYNC_REQUEST", 0);
     d->wmSyncRequestCounterAtom =
 	XInternAtom (dpy, "_NET_WM_SYNC_REQUEST_COUNTER", 0);
 
+    d->wmFullscreenMonitorsAtom =
+	XInternAtom (dpy, "_NET_WM_FULLSCREEN_MONITORS", 0);
+
     d->closeWindowAtom	    = XInternAtom (dpy, "_NET_CLOSE_WINDOW", 0);
     d->wmMoveResizeAtom	    = XInternAtom (dpy, "_NET_WM_MOVERESIZE", 0);
     d->moveResizeWindowAtom = XInternAtom (dpy, "_NET_MOVERESIZE_WINDOW", 0);
@@ -2171,10 +2097,6 @@ addDisplay (const char *name)
 
     d->toolkitActionAtom	  =
 	XInternAtom (dpy, "_COMPIZ_TOOLKIT_ACTION", 0);
-    d->toolkitActionMainMenuAtom  =
-	XInternAtom (dpy, "_COMPIZ_TOOLKIT_ACTION_MAIN_MENU", 0);
-    d->toolkitActionRunDialogAtom =
-	XInternAtom (dpy, "_COMPIZ_TOOLKIT_ACTION_RUN_DIALOG", 0);
     d->toolkitActionWindowMenuAtom  =
 	XInternAtom (dpy, "_COMPIZ_TOOLKIT_ACTION_WINDOW_MENU", 0);
     d->toolkitActionForceQuitDialogAtom  =
@@ -2394,34 +2316,14 @@ addDisplay (const char *name)
 
 	wmSnTimestamp = event.xproperty.time;
 
-	XSetSelectionOwner (dpy, wmSnAtom, newWmSnOwner, wmSnTimestamp);
-
-	if (XGetSelectionOwner (dpy, wmSnAtom) != newWmSnOwner)
+	if (!aquireSelection (d, i, "window", wmSnAtom, newWmSnOwner,
+			      wmSnTimestamp))
 	{
-	    compLogMessage ("core", CompLogLevelError,
-			    "Could not acquire window manager "
-			    "selection on screen %d display \"%s\"",
-			    i, DisplayString (dpy));
-
 	    XDestroyWindow (dpy, newWmSnOwner);
 
 	    continue;
 	}
 
-	/* Send client message indicating that we are now the WM */
-	event.xclient.type	   = ClientMessage;
-	event.xclient.window       = XRootWindow (dpy, i);
-	event.xclient.message_type = d->managerAtom;
-	event.xclient.format       = 32;
-	event.xclient.data.l[0]    = wmSnTimestamp;
-	event.xclient.data.l[1]    = wmSnAtom;
-	event.xclient.data.l[2]    = 0;
-	event.xclient.data.l[3]    = 0;
-	event.xclient.data.l[4]    = 0;
-
-	XSendEvent (dpy, XRootWindow (dpy, i), FALSE,
-		    StructureNotifyMask, &event);
-
 	/* Wait for old window manager to go away */
 	if (currentWmSnOwner != None)
 	{
@@ -2445,15 +2347,9 @@ addDisplay (const char *name)
 	    continue;
 	}
 
-	XSetSelectionOwner (dpy, cmSnAtom, newCmSnOwner, wmSnTimestamp);
-
-	if (XGetSelectionOwner (dpy, cmSnAtom) != newCmSnOwner)
+	if (!aquireSelection (d, i, "compositing", cmSnAtom,
+			      newCmSnOwner, wmSnTimestamp))
 	{
-	    compLogMessage ("core", CompLogLevelError,
-			    "Could not acquire compositing manager "
-			    "selection on screen %d display \"%s\"",
-			    i, DisplayString (dpy));
-
 	    continue;
 	}
 
@@ -2562,6 +2458,18 @@ removeDisplay (CompDisplay *d)
 
     objectFiniPlugins (&d->base);
 
+    if (d->edgeDelayHandle)
+    {
+	void *closure;
+
+	closure = compRemoveTimeout (d->edgeDelayHandle);
+	if (closure)
+	    free (closure);
+    }
+
+    if (d->autoRaiseHandle)
+	compRemoveTimeout (d->autoRaiseHandle);
+
     compRemoveTimeout (d->pingHandle);
 
     if (d->snDisplay)
@@ -2767,6 +2675,9 @@ handleSelectionRequest (CompDisplay *display,
 			     event->xselectionrequest.property,
 			     display->atomPairAtom,
 			     32, PropModeReplace, data, num);
+
+	    if (data)
+		XFree (data);
 	}
     }
     else
