@@ -124,61 +124,61 @@ handleSyncAlarm (CompWindow *w)
 static void
 moveInputFocusToOtherWindow (CompWindow *w)
 {
-    CompDisplay *display = w->screen->display;
+    CompScreen  *s = w->screen;
+    CompDisplay *d = s->display;
+    Bool        focussedAny = FALSE;
 
-    if (w->id == display->activeWindow)
+    if (w->id != d->activeWindow && w->id != d->nextActiveWindow)
+	return;
+
+    if (w->transientFor && w->transientFor != s->root)
     {
-	CompWindow *ancestor;
-
-	if (w->transientFor && w->transientFor != w->screen->root)
+	CompWindow *ancestor = findWindowAtDisplay (d, w->transientFor);
+	if (ancestor && !(ancestor->type & (CompWindowTypeDesktopMask |
+					    CompWindowTypeDockMask)))
 	{
-	    ancestor = findWindowAtDisplay (display, w->transientFor);
-	    if (ancestor && !(ancestor->type & (CompWindowTypeDesktopMask |
-						CompWindowTypeDockMask)))
-	    {
-		moveInputFocusToWindow (ancestor);
-	    }
-	    else
-		focusDefaultWindow (w->screen);
+	    moveInputFocusToWindow (ancestor);
+	    focussedAny = TRUE;
 	}
-	else if (w->type & (CompWindowTypeDialogMask |
-			    CompWindowTypeModalDialogMask))
-	{
-	    CompWindow *a, *focus = NULL;
-
-	    for (a = w->screen->reverseWindows; a; a = a->prev)
-	    {
-		if (a->clientLeader == w->clientLeader)
-		{
-		    if ((*w->screen->focusWindow) (a))
-		    {
-			if (focus)
-			{
-			    if (a->type & (CompWindowTypeNormalMask |
-					   CompWindowTypeDialogMask |
-					   CompWindowTypeModalDialogMask))
-			    {
-				if (compareWindowActiveness (focus, a) < 0)
-				    focus = a;
-			    }
-			}
-			else
-			    focus = a;
-		    }
-		}
-	    }
-
-	    if (focus && !(focus->type & (CompWindowTypeDesktopMask |
-					  CompWindowTypeDockMask)))
-	    {
-		moveInputFocusToWindow (focus);
-	    }
-	    else
-		focusDefaultWindow (w->screen);
-	}
-	else
-	    focusDefaultWindow (w->screen);
     }
+    else if (w->type & (CompWindowTypeDialogMask |
+			CompWindowTypeModalDialogMask))
+    {
+	CompWindow *a, *focus = NULL;
+
+	for (a = s->reverseWindows; a; a = a->prev)
+	{
+	    if (a->clientLeader != w->clientLeader)
+		continue;
+
+	    if (!(*s->focusWindow) (a))
+		continue;
+
+	    if (!focus)
+	    {
+		focus = a;
+		continue;
+	    }
+
+	    if (a->type & (CompWindowTypeNormalMask |
+			   CompWindowTypeDialogMask |
+			   CompWindowTypeModalDialogMask))
+	    {
+		if (compareWindowActiveness (focus, a) < 0)
+		    focus = a;
+	    }
+	}
+
+	if (focus && !(focus->type & (CompWindowTypeDesktopMask |
+				      CompWindowTypeDockMask)))
+	{
+	    moveInputFocusToWindow (focus);
+	    focussedAny = TRUE;
+	}
+    }
+
+    if (!focussedAny)
+	focusDefaultWindow (s);
 }
 
 static Bool
@@ -1380,8 +1380,9 @@ handleEvent (CompDisplay *d,
 		if (!w->attrib.override_redirect)
 		    setWmState (d, WithdrawnState, w->id);
 
-		w->placed  = FALSE;
-		w->managed = FALSE;
+		w->placed     = FALSE;
+		w->managed    = FALSE;
+		w->unmanaging = TRUE;
 	    }
 
 	    unmapWindow (w);
@@ -1433,7 +1434,8 @@ handleEvent (CompDisplay *d,
 
 		    if (w->id != d->activeWindow)
 			if (!(w->type & CompWindowTypeDockMask))
-			    moveInputFocusToWindow (w);
+			    if ((*s->focusWindow) (w))
+				moveInputFocusToWindow (w);
 		}
 	    }
 
@@ -1489,17 +1491,19 @@ handleEvent (CompDisplay *d,
 		state = getWindowState (d, w->id);
 		state = constrainWindowState (state, w->actions);
 
+		/* EWMH suggests that we ignore changes
+		   to _NET_WM_STATE_HIDDEN */
+		if (w->state & CompWindowStateHiddenMask)
+		    state |= CompWindowStateHiddenMask;
+		else
+		    state &= ~CompWindowStateHiddenMask;
+
 		if (state != w->state)
 		{
-		    w->state = state;
-
-		    recalcWindowType (w);
-		    recalcWindowActions (w);
-
 		    if (w->type & CompWindowTypeDesktopMask)
 			w->paint.opacity = OPAQUE;
 
-		    (*d->matchPropertyChanged) (d, w);
+		    changeWindowState (w, state);
 		}
 	    }
 	}
@@ -1646,9 +1650,10 @@ handleEvent (CompDisplay *d,
 
 		if (w->managed && w->startupId)
 		{
-		    Time         timestamp = 0;
-		    int          vx, vy, x, y;
-		    CompScreen   *s = w->screen;
+		    Time            timestamp = 0;
+		    int             vx, vy, x, y;
+		    CompScreen      *s = w->screen;
+		    CompFocusResult focus;
 
 		    w->initialTimestampSet = FALSE;
 		    applyStartupProperties (w->screen, w);
@@ -1665,13 +1670,13 @@ handleEvent (CompDisplay *d,
 		    y = w->attrib.y + (s->y - vy) * s->height;
 		    moveWindowToViewportPosition (w, x, y, TRUE);
 
-		    if (allowWindowFocus (w, 0,
-					  w->initialViewportX,
-					  w->initialViewportY,
-					  timestamp))
-		    {
+		    focus = allowWindowFocus (w, 0,
+					      w->initialViewportX,
+					      w->initialViewportY,
+					      timestamp);
+
+		    if (focus == CompFocusAllowed)
 			(*w->screen->activateWindow) (w);
-		    }
 		}
 	    }
 	}
@@ -1690,17 +1695,17 @@ handleEvent (CompDisplay *d,
 	    w = findWindowAtDisplay (d, event->xclient.window);
 	    if (w)
 	    {
-		Bool focusAllowed = TRUE;
+		CompFocusResult focus = CompFocusAllowed;
 
 		/* use focus stealing prevention if request came
 		   from an application */
 		if (event->xclient.data.l[0] == ClientTypeApplication)
-		    focusAllowed = allowWindowFocus (w, 0,
-						     w->screen->x,
-						     w->screen->y,
-						     event->xclient.data.l[1]);
+		    focus = allowWindowFocus (w, 0,
+					      w->screen->x,
+					      w->screen->y,
+					      event->xclient.data.l[1]);
 
-		if (focusAllowed)
+		if (focus == CompFocusAllowed)
 		    (*w->screen->activateWindow) (w);
 	    }
 	}
@@ -1781,11 +1786,12 @@ handleEvent (CompDisplay *d,
 
 		    /* raise the window whenever its fullscreen state,
 		       above/below state or maximization state changed */
-		    if (dState & (CompWindowStateFullscreenMask |
-				  CompWindowStateAboveMask |
-				  CompWindowStateBelowMask |
-				  CompWindowStateMaximizedHorzMask |
-				  CompWindowStateMaximizedVertMask))
+		    if (dState & CompWindowStateFullscreenMask)
+			stackingUpdateMode = CompStackingUpdateModeAboveFullscreen;
+		    else if (dState & (CompWindowStateAboveMask         |
+				       CompWindowStateBelowMask         |
+				       CompWindowStateMaximizedHorzMask |
+				       CompWindowStateMaximizedVertMask))
 			stackingUpdateMode = CompStackingUpdateModeNormal;
 
 		    changeWindowState (w, wState);
@@ -1968,6 +1974,21 @@ handleEvent (CompDisplay *d,
 	    if (w)
 		setDesktopForWindow (w, event->xclient.data.l[0]);
 	}
+	else if (event->xclient.message_type == d->wmFullscreenMonitorsAtom)
+	{
+	    w = findWindowAtDisplay (d, event->xclient.window);
+	    if (w)
+	    {
+		CompFullscreenMonitorSet monitors;
+
+		monitors.top    = event->xclient.data.l[0];
+		monitors.bottom = event->xclient.data.l[1];
+		monitors.left   = event->xclient.data.l[2];
+		monitors.right  = event->xclient.data.l[3];
+
+		setWindowFullscreenMonitors (w, &monitors);
+	    }
+	}
 	break;
     case MappingNotify:
 	updateModifierMappings (d);
@@ -1993,30 +2014,33 @@ handleEvent (CompDisplay *d,
 		}
 	    }
 
-	    w->managed = TRUE;
-
 	    if (w->state & CompWindowStateHiddenMask)
 		if (!w->minimized && !w->inShowDesktopMode)
 		    doMapProcessing = FALSE;
 
 	    if (doMapProcessing)
 	    {
-		Bool                   allowFocus;
-		CompStackingUpdateMode stackingMode;
-
 		w->initialViewportX = w->screen->x;
 		w->initialViewportY = w->screen->y;
 
 		w->initialTimestampSet = FALSE;
 
 		applyStartupProperties (w->screen, w);
+	    }
+
+	    w->managed = TRUE;
+
+	    if (doMapProcessing)
+	    {
+		CompFocusResult        focus;
+		CompStackingUpdateMode stackingMode;
 
 		if (!w->placed)
 		{
 		    int            newX, newY;
 		    int            gravity = w->sizeHints.win_gravity;
 		    XWindowChanges xwc;
-		    unsigned int   xwcm;
+		    unsigned int   xwcm, source;
 
 		    /* adjust for gravity */
 		    xwc.x      = w->serverX;
@@ -2026,28 +2050,30 @@ handleEvent (CompDisplay *d,
 
 		    xwcm = adjustConfigureRequestForGravity (w, &xwc,
 							     CWX | CWY,
-							     gravity);
+							     gravity, 1);
+
+		    source = ClientTypeApplication;
+		    (*w->screen->validateWindowResizeRequest) (w, &xwcm, &xwc,
+							       source);
+
+		    if (xwcm)
+			configureXWindow (w, xwcm, &xwc);
 
 		    if ((*w->screen->placeWindow) (w, xwc.x, xwc.y,
 						   &newX, &newY))
 		    {
 			xwc.x = newX;
 			xwc.y = newY;
-			xwcm |= CWX | CWY;
+			configureXWindow (w, CWX | CWY, &xwc);
 		    }
 
-		    if (xwcm)
-			configureXWindow (w, xwcm, &xwc);
-
 		    w->placed   = TRUE;
 		}
 
-		allowFocus = allowWindowFocus (w, NO_FOCUS_MASK,
-					       w->initialViewportX,
-					       w->initialViewportY,
-					       0);
+		focus = allowWindowFocus (w, NO_FOCUS_MASK,
+					  w->screen->x, w->screen->y, 0);
 
-		if (!allowFocus && (w->type & ~NO_FOCUS_MASK))
+		if (focus == CompFocusDenied)
 		    stackingMode = CompStackingUpdateModeInitialMapDeniedFocus;
 		else
 		    stackingMode = CompStackingUpdateModeInitialMap;
@@ -2059,13 +2085,13 @@ handleEvent (CompDisplay *d,
 
 		(*w->screen->leaveShowDesktopMode) (w->screen, w);
 
-		if (!(w->state & CompWindowStateHiddenMask))
-		{
-		    w->pendingMaps++;
-		    XMapWindow (d->display, w->id);
-		}
+		if (focus == CompFocusAllowed && !onCurrentDesktop (w))
+		    setCurrentDesktop (w->screen, w->desktop);
 
-		if (allowFocus)
+		if (!(w->state & CompWindowStateHiddenMask))
+		    showWindow (w);
+
+		if (focus == CompFocusAllowed)
 		    moveInputFocusToWindow (w);
 	    }
 
@@ -2095,8 +2121,9 @@ handleEvent (CompDisplay *d,
 
 	    if (event->xconfigurerequest.value_mask & CWStackMode)
 	    {
-		Window     above    = None;
-		CompWindow *sibling = NULL;
+		Window          above    = None;
+		CompWindow      *sibling = NULL;
+		CompFocusResult focus;
 
 		if (event->xconfigurerequest.value_mask & CWSibling)
 		{
@@ -2106,8 +2133,9 @@ handleEvent (CompDisplay *d,
 
 		switch (event->xconfigurerequest.detail) {
 		case Above:
-		    if (allowWindowFocus (w, NO_FOCUS_MASK,
-					  w->screen->x, w->screen->y, 0))
+		    focus = allowWindowFocus (w, NO_FOCUS_MASK,
+					      w->screen->x, w->screen->y, 0);
+		    if (focus == CompFocusAllowed)
 		    {
 			if (above)
 			{
@@ -2180,6 +2208,9 @@ handleEvent (CompDisplay *d,
 
 		state &= ~CompWindowStateDemandsAttentionMask;
 		changeWindowState (w, state);
+
+		if (d->nextActiveWindow == event->xfocus.window)
+		    d->nextActiveWindow = None;
 	    }
 	}
 	break;
@@ -2197,10 +2228,9 @@ handleEvent (CompDisplay *d,
 	    if (!d->opt[COMP_DISPLAY_OPTION_CLICK_TO_FOCUS].value.b &&
 		!s->maxGrab				            &&
 		event->xcrossing.mode   != NotifyGrab		    &&
-		event->xcrossing.mode   != NotifyUngrab		    &&
 		event->xcrossing.detail != NotifyInferior)
 	    {
-		Bool raise;
+		Bool raise, focus;
 		int  delay;
 
 		raise = d->opt[COMP_DISPLAY_OPTION_AUTORAISE].value.b;
@@ -2212,8 +2242,12 @@ handleEvent (CompDisplay *d,
 		    d->autoRaiseHandle = 0;
 		}
 
-		if (w->type & ~(CompWindowTypeDockMask |
-				CompWindowTypeDesktopMask))
+		if (w->type & NO_FOCUS_MASK)
+		    focus = FALSE;
+		else
+		    focus = (*w->screen->focusWindow) (w);
+
+		if (focus)
 		{
 		    moveInputFocusToWindow (w);
 
